from typing import Optional, Type, TypeVar, cast
from unittest import TestCase

from fileman.settings import SETTINGS, FilemanSettings, Settings


class TestSettings(Settings):
    """Settings manager for tests.

    Does not actually load settings from any configuration file.
    """

    def load(self, config_argv: Optional[str] = None) -> None:
        """Initiate an empty settings.

        Does nothing if called multiple times.
        """
        if self._settings is None:
            self._settings = FilemanSettings()


T = TypeVar('T', bound=Type[TestCase])


# XXX: Using snake_case looks much better for this context manager. It reminds override_settings from Django.
class override_settings:
    """Decorator and context manager to temporarily override settings."""

    def __init__(self, **overrides):
        self.overrides = overrides
        self.old_settings = None

    def enable(self):
        """Start settings override."""
        self.old_settings = SETTINGS.settings
        SETTINGS.settings = TestSettings()
        for key, value in dict(self.old_settings.items(), **self.overrides).items():
            SETTINGS[key] = value

    def disable(self):
        """Stop settings override."""
        SETTINGS.settings = cast(Settings, self.old_settings)

    def __enter__(self):
        self.enable()  # pragma: no cover

    def __exit__(self, exc_type, exc_value, traceback):
        self.disable()  # pragma: no cover

    def __call__(self, decorated: T) -> T:
        """Decorate TestCase."""
        original_setup = decorated.setUp
        original_teardown = decorated.tearDown

        def setUp(inner_self):
            self.enable()
            try:
                original_setup(inner_self)
            except Exception:  # pragma: no cover
                self.disable()
                raise

        def tearDown(inner_self):
            original_teardown(inner_self)
            self.disable()

        decorated.setUp = setUp  # type: ignore
        decorated.tearDown = tearDown  # type: ignore
        return decorated
